// ============================================================================
// COPYRIGHT NOTICE
// ----------------------------------------------------------------------------
// (This is the open source ISC license, see
// http://en.wikipedia.org/wiki/ISC_license
// for more info)
//
// Copyright © 2020-2023  Andreas M. Rammelt <rammi@caff.de>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//=============================================================================
// Latest version on https://caff.de/projects/decaff-commons/
//=============================================================================
package de.caff.generics;

import de.caff.annotation.NotNull;
import de.caff.generics.algorithm.DualPivotQuicksort;
import de.caff.generics.function.*;

import java.util.*;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 * Primitive short indexable with fixed length but mutable content.
 * This may be used as a substitute for {@code short[]} as it e.g. allows
 * read-only views or transparent views of subsets of the array.
 * @author <a href="mailto:rammi@caff.de">Rammi</a>
 * @since November 01, 2019
 */
public interface MutableShortIndexable
        extends ShortIndexable,
                Copyable<MutableShortIndexable>
{
  /**
   * Set the element at the given index.
   * @param index index between {@code 0} and {@code size() - 1}
   * @param value value to put to the given index
   */
  void set(int index, short value);

  /**
   * Pythonesque set.
   * This allows to access elements from the back by using negative indexes,
   * e.g. {@code -1} references the last element, {@code -2} its predecessor, and so on.
   * @param index index between {@code -size()} and {@code size() - 1}
   * @param value value to put to the given index
   */
  default void syt(int index, short value)
  {
    if (index >= 0) {
      set(index, value);
    }
    final int realIndex = index + size();
    if (realIndex < 0) {
      throw  new IndexOutOfBoundsException(String.format("Cannot access index %d with %d elements!",
                                                         index, size()));
    }
    set(realIndex, value);
  }

  /**
   * Set the values of this mutable short indexable one after
   * the other from the given Number iterable until either the
   * iterable is emptied or {@link #size()} elements are set.
   * @param iterable iterable from which this indexable is filled
   * @return number of elements set
   */
  default int setFrom(@NotNull Iterable<? extends Number> iterable)
  {
    return setFrom(iterable, 0, size());
  }

  /**
   * Set the values of this mutable short indexable one after
   * the other from the given iterable until either the
   * iterable is emptied or {@code numElements} elements are set.
   * @param iterable     iterable from which this indexable is filled
   * @param startIndex   start index where the setting begins
   * @param numElements  number of elements to set
   * @return number of elements set
   */
  default int setFrom(@NotNull Iterable<? extends Number> iterable,
                      int startIndex, int numElements)
  {
    if (startIndex > size() - numElements) {
      numElements = size() - startIndex;
    }

    int n = 0;

    for (Number elem : iterable) {
      set(startIndex + n++, elem.shortValue());
      if (n == numElements) {
        break;
      }
    }
    return n;
  }

  /**
   * Set consecutive entries from an array.
   * This will set less elements
   * @param array       array with elements to set from
   * @param arrayIndex  start index in the array
   * @param startIndex  start index in this mutable indexable
   * @param numElements number of elements to copy
   * @throws IndexOutOfBoundsException if this mutable indexable would overflow by this operation
   * @throws ArrayIndexOutOfBoundsException if array overflows by this operation
   */
  default void setFromArray(@NotNull short[] array,
                            int arrayIndex,
                            int startIndex,
                            int numElements)
  {
    if (startIndex + numElements > size()) {
      throw new IndexOutOfBoundsException("Overflow: numElements is too large!");
    }

    for (int n = 0;  n < numElements;  ++n) {
      set(startIndex + n, array[arrayIndex++]);
    }
  }

  /**
   * Set multiple elements to the same value.
   * @param from  first index to be set (Pythonesque)
   * @param len   number of elements to be set (non-negative)
   * @param value value to be set
   */
  default void setMulti(int from, int len, short value)
  {
    from = Pythonesque.mapX(from, this);
    if (len < 0) {
      throw new IllegalArgumentException("len has to be non-negative: "+len);
    }
    final int end = from + len;
    if (end > size()) {
      throw new IndexOutOfBoundsException(String.format("from + len exceed size(): %d + %d > %d", from, len, size()));
    }
    for (int i = from;  i < end;  ++i) {
      set(i, value);
    }
  }

  /**
   * Copy elements inside this indexable.
   * @param fromIndex   start index from where elements are copied (Pythonesque)
   * @param toIndex     start index to where elements are copied (Pythonesque)
   * @param numElements number of elements to copy
   */
  default void copyInternally(int fromIndex, int toIndex, int numElements)
  {
    if (numElements == 0) {
      return;
    }
    if (numElements < 0) {
      throw new IllegalArgumentException("numElements must not be negative!");
    }
    final int size = size();
    int from = Pythonesque.mapX(fromIndex, size);
    int to   = Pythonesque.mapX(toIndex, size);
    if (from + numElements > size) {
      throw new IndexOutOfBoundsException(String.format("Reading would overflow: from + numElements > size (%d + %d > %d)!",
                                                        from, numElements, size));
    }
    if (to + numElements > size) {
      throw new IndexOutOfBoundsException(String.format("Writing would overflow: to + numElements > size (%d + %d > %d)!",
                                                        to, numElements, size));
    }
    if (from == to) {
      return;
    }
    if (from < to  &&  from + numElements > to) {
      // have to go backward
      from += numElements;
      to   += numElements;
      while (--numElements >= 0) {
        set(--to, get(--from));
      }
    }
    else {
      while (--numElements >= 0) {
        set(to++, get(from++));
      }
    }
  }

  /**
   * Fill this indexable from a provider.
   * @param provider provider which is called for each index and provides the element for that index
   */
  default void fillFrom(@NotNull IntToShortFunction1 provider)
  {
    final int size = size();
    for (int p = 0;  p < size;  ++p) {
      set(p, provider.applyAsShort(p));
    }
  }

  /**
   * Swap the values at two indices.
   * @param idx1 first index
   * @param idx2 second index
   */
  default void swap(int idx1, int idx2)
  {
    if (idx1 == idx2) {
      return;
    }
    final short tmp = get(idx1);
    set(idx1, get(idx2));
    set(idx2, tmp);
  }

  /**
   * Swap the values at two indices using Pythonesque indices.
   * @param idx1 first index ({@link Pythonesque})
   * @param idx2 second index ({@link Pythonesque})
   */
  default void swyp(int idx1, int idx2)
  {
    if (idx1 == idx2) {
      return;
    }
    swap(Pythonesque.mapX(idx1, this),
         Pythonesque.mapX(idx2, this));
  }

  @NotNull
  @Override
  default Base subSet(int fromIndex, int toIndex)
  {
    if (fromIndex < 0) {
      throw new IndexOutOfBoundsException("fromIndex = " + fromIndex);
    }
    if (toIndex > size()) {
      throw new IndexOutOfBoundsException("toIndex = " + toIndex);
    }
    if (fromIndex > toIndex) {
      throw new IllegalArgumentException("fromIndex(" + fromIndex +
                                         ") > toIndex(" + toIndex + ")");
    }
    final int length = toIndex - fromIndex;

    return new Base()
    {
      @Override
      public void set(int index, short elem)
      {
        if (index < 0  ||  index >= length) {
          throw new IndexOutOfBoundsException("index = "+index);
        }
        MutableShortIndexable.this.set(index + fromIndex,
                                       elem);
      }

      @Override
      public int size()
      {
        return length;
      }

      @Override
      public short get(int index)
      {
        if (index < 0  ||  index >= length) {
          throw new IndexOutOfBoundsException("index = "+index);
        }
        return MutableShortIndexable.this.get(index + fromIndex);
      }

      @NotNull
      @Override
      public MutableShortIndexable.Base subSet(int fromIdx, int toIdx)
      {
        if (fromIdx < 0) {
          throw new IndexOutOfBoundsException("fromIndex = " + fromIdx);
        }
        if (toIdx > size()) {
          throw new IndexOutOfBoundsException("toIndex = " + toIdx);
        }
        if (fromIdx > toIdx) {
          throw new IllegalArgumentException("fromIndex(" + fromIdx +
                                             ") > toIndex(" + toIdx + ")");
        }
        return MutableShortIndexable.this.subSet(fromIndex + fromIdx,
                                                  fromIndex + toIdx - fromIdx);
      }

      @Override
      public void copyInternally(int fromIdx, int toIdx, int numElements)
      {
        final int from = Pythonesque.mapX(fromIndex, length);
        final int to   = Pythonesque.mapX(toIndex, length);
        if (from + numElements > length) {
          throw new IndexOutOfBoundsException(String.format("Reading would overflow: from + numElements > size (%d + %d > %d)!",
                                                            from, numElements, length));
        }
        if (to + numElements > length) {
          throw new IndexOutOfBoundsException(String.format("Writing would overflow: to + numElements > size (%d + %d > %d)!",
                                                            to, numElements, length));
        }
        // now forward to wrapped
        MutableShortIndexable.this.copyInternally(fromIndex + fromIdx,
                                                  fromIndex + toIdx,
                                                  numElements);
      }
    };
  }

  @NotNull
  @Override
  default Base sybSet(int fromIndex, int toIndex)
  {
    return subSet(fromIndex < 0
                          ? size() + fromIndex
                          : fromIndex,
                  toIndex < 0
                          ? size() + toIndex
                          : toIndex);
  }

  @NotNull
  @Override
  default Base tailSet(int fromIndex)
  {
    return subSet(fromIndex < 0
                          ? size() + fromIndex
                          : fromIndex,
                  size());
  }

  @NotNull
  @Override
  default Base headSet(int toIndex)
  {
    return subSet(0,
                  toIndex < 0
                          ? size() + toIndex
                          : toIndex);
  }

  @NotNull
  @Override
  default Base reverse()
  {
    return new MutableShortIndexable.Base()
    {
      @Override
      public void set(int index, short elem)
      {
        MutableShortIndexable.this.set(size() - index - 1, elem);
      }

      @Override
      public int size()
      {
        return MutableShortIndexable.this.size();
      }

      @Override
      public short get(int index)
      {
        return MutableShortIndexable.this.get(size() - index - 1);
      }

      @NotNull
      @Override
      public MutableShortIndexable.Base reverse()
      {
        return based(MutableShortIndexable.this);
      }
    };
  }

  /**
   * Set all values of this short indexable by index.
   * @param setter setter which provides the value to be set for a given index
   */
  default void initByIndex(@NotNull IntToShortFunction1 setter)
  {
    final int size = size();
    for (int i = 0; i < size; ++i) {
      set(i, setter.applyAsShort(i));
    }
  }

  /**
   * Revert the order of all elements in the given range.
   * Attention: the {@code to} element is also included in the reversion to allow easy
   * reversion at the end by using {@code -1}.
   * @param from start element of reversion (Pythonesque), included
   * @param to   end element of reversion (Pythonesque), included
   */
  default void revert(int from, int to)
  {
    final int size = size();
    int up = Pythonesque.map(from, size);
    int down = Pythonesque.map(to, size);

    while (up < down) {
      swap(up++, down--);
    }
  }

  /**
   * Revert the elements in this indexable.
   * This reverts the order of all elements.
   */
  default void revert()
  {
    if (size() <= 1) {
      return;
    }
    revert(0, -1);
  }

  /**
   * Sort the elements in this indexable in their natural order.
   */
  default void order()
  {
    final short[] array = toArray();
    Arrays.sort(array);
    fillFrom(idx -> array[idx]);
  }

  /**
   * Sort the elements in this indexable according to the given ordering.
   * @param ordering sort order
   */
  default void order(@NotNull ShortOrdering ordering)
  {
    final short[] array = toArray();
    DualPivotQuicksort.sort(array, ordering);
    fillFrom(idx -> array[idx]);
  }

  /**
   * Randomize the content  of this mutable indexable.
   * This will reorder the elements randomly.
   * @param random random number generator
   */
  default void shuffle(@NotNull Random random)
  {
    final int size = size();
    for (int i = size - 1;  i >= 0;  --i) {
      swap(i, random.nextInt(i + 1));
    }
  }

  /**
   * Convert this mutable short array into an integer array.
   * <p>
   * It is recommended to do this when this indexable is heavily used.
   *
   * @return mutable int indexable copy of this short indexable
   */
  @NotNull
  default MutableIntIndexable.Base toMutableIntIndexable()
  {
    return MutableIntIndexable.viewArray(toIntArray());
  }

  /**
   * View this indexable as a standard list.
   * Overwritten to allow for modifications.
   * The returned list will allow usage of the {@link List#set(int, Object)} method,
   * but neither any adding nor deleting methods. Standard sorting algorithms will work
   * on the returned list as a sorting algorithm is expected to neither move nor add
   * elements.
   * @return list allowing setting elements
   */
  @NotNull
  @Override
  default List<Short> asList()
  {
    return new AbstractList<Short>()
    {
      @Override
      public Short get(int index)
      {
        return MutableShortIndexable.this.get(index);
      }

      @Override
      public int size()
      {
        return MutableShortIndexable.this.size();
      }

      @Override
      public Short set(int index, Short element)
      {
        MutableShortIndexable.this.set(index, element);
        return element;
      }

      @Override
      public Iterator<Short> iterator()
      {
        return MutableShortIndexable.this.iterator();
      }
    };
  }

  @NotNull
  @Override
  default MutableShortIndexable getCopy()
  {
    return viewArray(toArray());
  }

  /**
   * View any mutable short indexable as a base short indexable.
   * The latter provides various standard methods for comparison, output and hanshing.
   * @param indexable mutqable short indexable to view as a {@link Base} mutable short indexable
   * @return base view of given indexable, or the indexable itself if it is already a Base
   */
  @NotNull
  static Base based(@NotNull MutableShortIndexable indexable)
  {
    if (indexable instanceof Base) {
      return (Base)indexable;
    }
    return new Base()
    {
      @Override
      public void set(int index, short value)
      {
        indexable.set(index, value);
      }

      @Override
      public int size()
      {
        return indexable.size();
      }

      @Override
      public short get(int index)
      {
        return indexable.get(index);
      }
    };
  }

  /**
   * Initialize a mutable indexable to a given size.
   * @param size     size of indexable
   * @param creator  creator for the initial elements
   * @return mutable indexable with the given size
   * @see #zeroed(int)
   * @see #init(int, short)
   */
  @NotNull
  static Base init(int size, @NotNull Supplier<? extends Number> creator)
  {
    final short[] array = new short[size];
    for (int i = 0;  i < size;  ++i) {
      array[i] = creator.get().shortValue();
    }
    return viewArray(array);
  }

  /**
   * Get a mutable short indexable of a given size which contains elements created by index.
   * @param size size of the returned indexable
   * @param producer producer which is called with an index and expected to return the associated value              
   * @return a mutable short indexable with the given size with elements initialized by calling the {@code producer}
   */
  @NotNull
  static Base initByIndex(int size, @NotNull IntToShortFunction1 producer)
  {
    if (size == 0) {
      return EMPTY;
    }
    if (size < 0) {
      throw new IndexOutOfBoundsException("Indexables with negative size are impossible: "+size);
    }
    final short[] array = new short[size];
    for (int i = 0;  i < size;  ++i) {
      array[i] = producer.applyAsShort(i);
    }
    return viewArray(array);
  }

  /**
   * Create a mutable indexable which is initialized from a
   * given collection.
   * @param collection collection
   * @return mutable indexable
   */
  @NotNull
  static Base copyOf(@NotNull Collection<? extends Number> collection)
  {
    final short[] array = new short[collection.size()];
    int i = 0;
    for (Number n : collection) {
      array[i++] = n.shortValue();
    }
    return viewArray(array);
  }

  /**
   * Create a mutable indexable which is initialized from copied elements of
   * a given collection.
   * @param collection collection
   * @param copier     element copier
   * @param <IN>       incoming element type
   * @return mutable indexable
   */
  @NotNull
  static <IN> Base copy(@NotNull Collection<IN> collection,
                        @NotNull Function<IN, Number> copier)
  {
    final short[] array = new short[collection.size()];
    int i = 0;
    for (IN elem : collection) {
      array[i++] = copier.apply(elem).shortValue();
    }
    return viewArray(array);
  }

  @NotNull
  static Base empty()
  {
    return EMPTY;
  }

  /**
   * Create a mutable indexable which is initialized from elements
   * of the given array.
   * @param elements  elements
   * @return mutable indexable independent of {@code elements}
   */
  @NotNull
  static Base fromArray(@NotNull short... elements)
  {
    return fromArray(elements, 0, elements.length);
  }

  /**
   * Create a mutable indexable which is initialized from elements
   * of the given array.
   * @param elements    array of elements
   * @param startIndex  index of first element used in the returned indxable
   * @param length      length number of elements used in the returned indexable
   * @return mutable indexable independent from {@code elements}
   */
  @NotNull
  static Base fromArray(@NotNull short[] elements,
                        int startIndex, int length)
  {
    if (startIndex < 0  ||  startIndex + length > elements.length)  {
      throw new IllegalArgumentException("Indexes out of bounds!");
    }
    return viewArray(Arrays.copyOfRange(elements, startIndex, startIndex + length));
  }

  /**
   * Mutable indexable view which operates on the given array.
   * Any changes done by the returned indexable will be forwarded to the array.
   * @param array array
   * @return mutable indexable view of the given array
   */
  @NotNull
  static Base viewArray(@NotNull short[] array)
  {
    return new Base()
    {
      @Override
      public void set(int index, short elem)
      {
        array[index] = elem;
      }

      @Override
      public int size()
      {
        return array.length;
      }

      @Override
      public short get(int index)
      {
        return array[index];
      }

      @NotNull
      @Override
      public MutableShortIndexable.Base subSet(int fromIndex, int toIndex)
      {
        return MutableShortIndexable.viewArray(array, fromIndex, toIndex - fromIndex);
      }

      @Override
      public void copyInternally(int fromIndex, int toIndex, int numElements)
      {
        System.arraycopy(array, fromIndex, array, toIndex, numElements);
      }

      @Override
      public void order()
      {
        Arrays.sort(array);
      }

      @Override
      public void order(@NotNull ShortOrdering ordering)
      {
        DualPivotQuicksort.sort(array, ordering);
      }
    };
  }

  /**
   * Mutable indexable view which operates on a part of the given array.
   * @param array  array
   * @param start  start position in array
   * @param length length of returned indexable
   * @return mutable short indexable view of the given part of the array
   */
  @NotNull
  static MutableShortIndexable.Base viewArray(@NotNull short[] array, int start, int length)
  {
    if (start < 0) {
      throw new IndexOutOfBoundsException("start has to be non-negative, but is "+start);
    }
    if (length < 0) {
      throw new IllegalArgumentException("len has to be non-negative, but is "+length);
    }
    if (start + length > array.length) {
      throw new IndexOutOfBoundsException(String.format("end will be outside array: %d + %d > %d",
                                                        start, length, array.length));
    }
    if (length == 0) {
      return empty();
    }
    return new MutableShortIndexable.Base()
    {
      @Override
      public void set(int index, short value)
      {
        array[index + start] = value;
      }

      @Override
      public short get(int index)
      {
        return array[index + start];
      }

      @Override
      public int size()
      {
        return length;
      }

      @NotNull
      @Override
      public MutableShortIndexable.Base subSet(int fromIndex, int toIndex)
      {
        if (fromIndex < 0) {
          throw new IndexOutOfBoundsException("fromIndex = " + fromIndex);
        }
        if (toIndex > length) {
          throw new IndexOutOfBoundsException("toIndex = " + toIndex);
        }
        if (fromIndex > toIndex) {
          throw new IllegalArgumentException("fromIndex(" + fromIndex +
                                             ") > toIndex(" + toIndex + ")");
        }
        return MutableShortIndexable.viewArray(array,
                                                fromIndex + start,
                                                toIndex - fromIndex);
      }

      @Override
      public void copyInternally(int fromIndex, int toIndex, int numElements)
      {
        if (numElements == 0) {
          return;
        }
        if (numElements < 0) {
          throw new IllegalArgumentException("numElements must not be negative!");
        }
        final int from = Pythonesque.mapX(fromIndex, length);
        final int to   = Pythonesque.mapX(toIndex, length);
        if (from + numElements > length) {
          throw new IndexOutOfBoundsException(String.format("Reading would overflow: from + numElements > size (%d + %d > %d)!",
                                                            from, numElements, length));
        }
        if (to + numElements > length) {
          throw new IndexOutOfBoundsException(String.format("Writing would overflow: to + numElements > size (%d + %d > %d)!",
                                                            to, numElements, length));
        }
        if (from == to) {
          return;
        }
        System.arraycopy(array, fromIndex + start, array, toIndex + start, numElements);
      }

      @Override
      public void order()
      {
        Arrays.sort(array, start, start + length);
      }

      @Override
      public void order(@NotNull ShortOrdering ordering)
      {
        DualPivotQuicksort.sort(array, start, start + length - 1, ordering);
      }
    };
  }

  /**
   * Mutable indexable view which operates on the given array.
   * Any changes done by the returned indexable will be forwarded to the list.
   * @param list list
   * @return mutable indexable view of the given list
   */
  @NotNull
  static Base viewList(@NotNull List<Short> list)
  {
    return new Base()
    {
      @Override
      public void set(int index, short elem)
      {
        list.set(index, elem);
      }

      @Override
      public int size()
      {
        return list.size();
      }

      @Override
      public short get(int index)
      {
        return list.get(index);
      }
    };
  }

  /**
   * Create a mutable indexable which is the copy of a standard indexable.
   * @param indexable standard indexable
   * @return mutable indexable initialize from the standard indexable
   */
  @NotNull
  static Base fromIndexable(@NotNull Indexable<? extends Number> indexable)
  {
    return copyOf(indexable.asCollection());
  }

  /**
   * Create a mutable indexable which is the copy of a standard indexable.
   * This especieally allows to copy each element during the initialization.
   * @param indexable  base indexable
   * @param converter  converter from base indexable type to result element type,
   * @param <B> element type of incoming idexable
   * @return mutable indexable
   */
  @NotNull
  static <B> Base fromIndexable(@NotNull Indexable<B> indexable,
                                @NotNull Function<? super B, ? extends Number> converter)
  {
    final short[] array = new short[indexable.size()];
    int i = 0;
    for (B elem : indexable) {
      array[i++] = converter.apply(elem).shortValue();
    }
    return viewArray(array);
  }

  /**
   * View a generic indexable as a mutable short indexable by accessing
   * a short property of the elements of the underlying generic indexable.
   * <p>
   * This is useful if you have complex items, but are interested
   * into only one (short) property of each item.
   * @param indexable underlying generic indexable
   * @param getter    function used to extract the property of interest
   * @param setter    procedure used to set the property of interest
   * @param <T>       element type of the underlying indexable
   * @return mutable short indexable view of the underlying indexable
   */
  @NotNull
  static <T> MutableShortIndexable.Base viewIndexable(@NotNull Indexable<T> indexable,
                                                      @NotNull ToShortFunction<? super T> getter,
                                                      @NotNull ShortSetter<? super T> setter)
  {
    return new MutableShortIndexable.Base()
    {
      @Override
      public void set(int index, short value)
      {
        setter.set(indexable.get(index), value);
      }

      @Override
      public short get(int index)
      {
        return getter.applyAsShort(indexable.get(index));
      }

      @Override
      public int size()
      {
        return indexable.size();
      }
    };
  }

  /**
   * Return a mutable short indexable initialized with the content of a
   * standard double indexable. This will cast the incoming double
   * values and result in a loss of accuracy.
   * @param indexable double indexable
   * @return mutable indexable initialized with the values of the given double indexable
   */
  @NotNull
  static Base fromDoubleIndexable(@NotNull DoubleIndexable indexable)
  {
    return initByIndex(indexable.size(), i -> (short)indexable.get(i));
  }

  /**
   * Return a mutable short indexable initialized with the content of a
   * standard float indexable. This will cast the incoming float
   * values and result in a loss of accuracy.
   * @param indexable float indexable
   * @return mutable indexable initialized with the values of the given float indexable
   */
  @NotNull
  static Base fromFloatIndexable(@NotNull FloatIndexable indexable)
  {
    return initByIndex(indexable.size(), i -> (short)indexable.get(i));
  }

  /**
   * Return a mutable short indexable initialized with the content of a
   * standard long indexable. This will cast long values to short.
   * @param indexable long indexable
   * @return mutable indexable initialized with the values of the given long indexable
   */
  @NotNull
  static Base fromLongIndexable(@NotNull LongIndexable indexable)
  {
    return initByIndex(indexable.size(), i -> (short)indexable.get(i));
  }

  /**
   * Return a mutable short indexable initialized with the content of a
   * standard int indexable. This will result in a loss of accuracy for
   * various values.
   * @param indexable int indexable
   * @return mutable indexable initialized with the values of the given int indexable
   */
  @NotNull
  static Base fromIntIndexable(@NotNull IntIndexable indexable)
  {
    return initByIndex(indexable.size(), i -> (short)indexable.get(i));
  }

  /**
   * Return a mutable short indexable initialized with the content of a
   * standard short indexable.
   * @param indexable short indexable
   * @return mutable indexable initialized with the values of the given short indexable
   */
  @NotNull
  static Base fromShortIndexable(@NotNull ShortIndexable indexable)
  {
    return initByIndex(indexable.size(), indexable::get);
  }

  /**
   * Return a mutable short indexable initialized with the content of a
   * standard byte indexable.
   * @param indexable long indexable
   * @return mutable indexable initialized with the values of the given byte indexable
   */
  @NotNull
  static Base fromByteIndexable(@NotNull ByteIndexable indexable)
  {
    return initByIndex(indexable.size(), indexable::get);
  }

  /**
   * Create a mutable indexable from an iterable or a part of it.
   * @param size maximum size of returned indexable
   * @param iter iterable used for initializing the indexable
   * @return mutable indexable
   */
  @NotNull
  static Base fromIterable(int size, @NotNull Iterable<? extends Number> iter)
  {
    final short[] array = new short[size];
    int i = 0;
    for (Number elem : iter) {
      array[i++] = elem.shortValue();
      if (i == size) {
        break;
      }
    }
    return viewArray(array);
  }

  /**
   * Create a mutable short indexable with {@code size}
   * {@code 0.0} elements.
   * @param size required size of returned mutable short indexable
   * @return mutable short indexable of the given size initialized with
   *         {@code 0.0}
   * @see #init(int, short)
   * @see #init(int, Supplier)
   */
  @NotNull
  static Base zeroed(int size)
  {
    return viewArray(new short[size]);
  }

  /**
   * Create a mutable
   * @param size  required size of returned mutable short indexable
   * @param value initial value of all elements
   * @return mutable short indexable of the given size initialized with
   *         {@code value}
   * @see #init(int, Supplier)
   * @see #zeroed(int)
   */
  @NotNull
  static Base init(int size, short value)
  {
    final short[] array = new short[size];
    Arrays.fill(array, value);
    return viewArray(array);
  }

  /**
   * Abstract base class which provides useful implementations
   * for {@link Object#equals(Object)}, {@link Object#hashCode()},
   * {@link Object#toString()}.
   * @see Indexable#asBase()
   */
  abstract class Base extends ShortIndexable.Base implements MutableShortIndexable
  {
  }

  /**
   * Empty mutable indexable.
   * Use {@link #empty()} instead.
   */
  Base EMPTY = new MutableShortIndexable.Base()
  {
    @Override
    public void set(int index, short value)
    {
      throw new IndexOutOfBoundsException("Cannot set in empty indexable!");
    }

    @Override
    public int size()
    {
      return 0;
    }

    @Override
    public short get(int index)
    {
      throw new IndexOutOfBoundsException("Cannot get from empty indexable!");
    }

    @NotNull
    @Override
    public ListIterator<Short> listIterator()
    {
      return Types.emptyListIterator();
    }

    @NotNull
    @Override
    public MutableShortIndexable.Base reverse()
    {
      return this;
    }

    @Override
    public boolean isEmpty()
    {
      return true;
    }

    @NotNull
    @Override
    public Iterator<Short> iterator()
    {
      return Types.emptyIterator();
    }

    @NotNull
    @Override
    public PrimitiveShortIterator shortIterator()
    {
      return Types.EMPTY_SHORT_ITERATOR;
    }

    @NotNull
    @Override
    public PrimitiveIterator.OfInt intIterator()
    {
      return Types.EMPTY_INT_ITERATOR;
    }

    @NotNull
    @Override
    public Collection<Short> asCollection()
    {
      return Collections.emptyList();
    }

    @NotNull
    @Override
    public List<Short> asList()
    {
      return Collections.emptyList();
    }

    @NotNull
    @Override
    public Iterable<Integer> indexes()
    {
      return Types.emptyIterable();
    }

    @NotNull
    @Override
    public IntIndexable intIndexes()
    {
      return IntIndexable.EMPTY;
    }

    @NotNull
    @Override
    public short[] toArray()
    {
      return Empty.SHORT_ARRAY;
    }

    @Override
    public void initByIndex(@NotNull IntToShortFunction1 setter)
    {
    }

    @Override
    public void forEachShort(@NotNull ShortConsumer action)
    {
    }

    @Override
    public <E extends Exception> void forEachShortFragile(@NotNull FragileShortConsumer<E> action) throws E
    {
    }

    @Override
    public boolean containsShort(short value)
    {
      return false;
    }

    @Override
    public int addToArray(@NotNull short[] array, int pos)
    {
      return pos;
    }

    @NotNull
    @Override
    public IntIndexable.Base asIntIndexable()
    {
      return IntIndexable.EMPTY;
    }

    @NotNull
    @Override
    public IntIndexable.Base asUnsignedIndexable()
    {
      return IntIndexable.EMPTY;
    }

    @Override
    public short foldLeft(short initialValue, @NotNull ShortOperator2 foldOperation)
    {
      return initialValue;
    }

    @NotNull
    @Override
    public ShortIndexable frozen()
    {
      return this;
    }

    @NotNull
    @Override
    public MutableShortIndexable getCopy()
    {
      return this;
    }

    @Override
    public void forEach(Consumer<? super Short> action)
    {
    }

    @Override
    public void order()
    {
    }

    @Override
    @NotNull
    public String toString()
    {
      return Indexable.EMPTY_INDEXABLE_STRING;
    }

    @Override
    public int hashCode()
    {
      return 1;
    }
  };
}
