// ============================================================================
// COPYRIGHT NOTICE
// ----------------------------------------------------------------------------
// (This is the open source ISC license, see
// http://en.wikipedia.org/wiki/ISC_license
// for more info)
//
// Copyright © 2019-2023  Andreas M. Rammelt <rammi@caff.de>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//=============================================================================
// Latest version on https://caff.de/projects/decaff-commons/
//=============================================================================
package de.caff.generics.function;

import de.caff.annotation.NotNull;

/**
 * Unary operator with a boolean argument and result.
 * @author <a href="mailto:rammi@caff.de">Rammi</a>
 * @since October 31, 2019
 */
@FunctionalInterface
public interface BooleanOperator1
{
  /** The standard boolean NOT ({@code !}) operator. */
  @NotNull
  BooleanOperator1 NOT = x -> !x;

  /**
   * Apply the function.
   * @param value parameter value of the function
   * @return result of the function
   */
  boolean applyAsBoolean(boolean value);

  /**
   * Create a zero argument function from this one by applying a constant argument.
   * Although useless for pure functions it can be useful in cases where the zero
   * argument function is a factory (and therefore not pure).
   * @param argument constant argument
   * @return zero argument function
   */
  @NotNull
  default BooleanOperator0 partial(boolean argument)
  {
    return () -> applyAsBoolean(argument);
  }

  /**
   * Create a function which applies this function and feeds the result into the next function.
   * @param after function called with the result of this function
   * @return chained function
   */
  @NotNull
  default BooleanOperator1 andThen(@NotNull BooleanOperator1 after)
  {
    return (arg) -> after.applyAsBoolean(applyAsBoolean(arg));
  }
}
