// ============================================================================
// COPYRIGHT NOTICE
// ----------------------------------------------------------------------------
// (This is the open source ISC license, see
// http://en.wikipedia.org/wiki/ISC_license
// for more info)
//
// Copyright © 2012-2023  Andreas M. Rammelt <rammi@caff.de>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//=============================================================================
// Latest version on https://caff.de/projects/decaff-commons/
//=============================================================================
package de.caff.generics.mda;

import de.caff.annotation.NotNull;

/**
 * Read access for 3-dimensional rectangular boolean arrays.
 * @author <a href="mailto:rammi@caff.de">Rammi</a>
 * @since November 11, 2019
 */
public interface ThreeDimensionalBooleanReadAccess
        extends ThreeDimensionalReadAccess<Boolean>
{
  /** Empty 3dimensional boolean read access. */
  ThreeDimensionalBooleanReadAccess EMPTY = new ThreeDimensionalBooleanReadAccess()
  {
    @Override
    public boolean getValueAt(int ix, int iy, int iz)
    {
      throw new IndexOutOfBoundsException();
    }

    @Override
    public int sizeX()
    {
      return 0;
    }

    @Override
    public int sizeY()
    {
      return 0;
    }

    @Override
    public int sizeZ()
    {
      return 0;
    }
  };

  /**
   * Get a 3dimensional boolean read access for a single element.
   * @param singleElement single element
   * @return 3dimensional boolean read access with size 1 in all 3 directions,
   *         containing just the given single element
   */
  @NotNull
  static ThreeDimensionalBooleanReadAccess singleton(boolean singleElement)
  {
    return new ThreeDimensionalBooleanReadAccess()
    {
      @Override
      public boolean getValueAt(int ix, int iy, int iz)
      {
        if (ix != 0  ||  iy != 0  ||  iz != 0) {
          throw new IllegalArgumentException();
        }
        return singleElement;
      }

      @Override
      public int sizeX()
      {
        return 1;
      }

      @Override
      public int sizeY()
      {
        return 1;
      }

      @Override
      public int sizeZ()
      {
        return 1;
      }
    };
  }

  /**
   * Get the value at the given indexes.
   * @param ix X index
   * @param iy Y index
   * @param iz Z index
   * @return value at the given index
   */
  boolean getValueAt(int ix, int iy, int iz);

  @Override
  @NotNull
  default Boolean getElementAt(int ix, int iy, int iz)
  {
    return getValueAt(ix, iy, iz);
  }

  @NotNull
  @Override
  default TwoDimensionalBooleanReadAccess subAtX(int fixX)
  {
    return new TwoDimensionalBooleanReadAccess()
    {
      @Override
      public boolean getValueAt(int ix, int iy)
      {
        return ThreeDimensionalBooleanReadAccess.this.getValueAt(fixX, ix, iy);
      }

      @Override
      public int sizeX()
      {
        return ThreeDimensionalBooleanReadAccess.this.sizeY();
      }

      @Override
      public int sizeY()
      {
        return ThreeDimensionalBooleanReadAccess.this.sizeZ();
      }
    };
  }

  @NotNull
  @Override
  default TwoDimensionalBooleanReadAccess subAtY(int fixY)
  {
    return new TwoDimensionalBooleanReadAccess()
    {
      @Override
      public boolean getValueAt(int ix, int iy)
      {
        return ThreeDimensionalBooleanReadAccess.this.getValueAt(ix, fixY, iy);
      }

      @Override
      public int sizeX()
      {
        return ThreeDimensionalBooleanReadAccess.this.sizeX();
      }

      @Override
      public int sizeY()
      {
        return ThreeDimensionalBooleanReadAccess.this.sizeZ();
      }
    };
  }

  @NotNull
  @Override
  default TwoDimensionalBooleanReadAccess subAtZ(int fixZ)
  {
    return new TwoDimensionalBooleanReadAccess()
    {
      @Override
      public boolean getValueAt(int ix, int iy)
      {
        return ThreeDimensionalBooleanReadAccess.this.getValueAt(ix, iy, fixZ);
      }

      @Override
      public int sizeX()
      {
        return ThreeDimensionalBooleanReadAccess.this.sizeX();
      }

      @Override
      public int sizeY()
      {
        return ThreeDimensionalBooleanReadAccess.this.sizeY();
      }
    };
  }

}
